from __future__ import annotations

import bz2
import logging
from pathlib import Path
from shutil import get_unpack_formats, register_unpack_format

logger = logging.getLogger(__name__)


def _unpack_bz2(filename, extract_dir):
    filename = Path(filename)
    extract_dir = Path(extract_dir)

    extracted = extract_dir / filename.stem

    with filename.open("rb") as filename_fd:
        with extracted.open("wb") as extracted_fd:
            extracted_fd.write(bz2.decompress(filename_fd.read()))


def register_unpack_formats():
    """Register custom unpack formats."""
    logger.debug("Registering custom unpack formats")

    formats = get_unpack_formats()
    if "bz2" not in map(lambda x: x[0], formats):
        register_unpack_format("bz2", [".bz2"], _unpack_bz2, description="bz2 files")

    logger.debug("Unpack formats available: %s", formats)



from __future__ import annotations

import hashlib
import logging
import re
from pathlib import Path

__all__ = [
    "compute_file_checksum",
    "find_checksum_in_file",
    "HASH_ALGORITHM",
    "is_hexdigest",
    "parse_checksum_option",
]

logger = logging.getLogger(__name__)

HASH_ALGORITHM = ("md5", "sha1", "sha224", "sha256", "sha384", "sha512")
HASH_ALGORITHM_LENGTH = {
    "md5": 32,
    "sha1": 40,
    "sha224": 56,
    "sha256": 64,
    "sha384": 96,
    "sha512": 128,
}


def parse_checksum_option(value: str) -> tuple[str, str]:
    try:
        algorithm, checksum = value.split(":", maxsplit=1)
    except ValueError as exception:
        raise ValueError(f"invalid checksum option {value}") from exception

    if algorithm not in HASH_ALGORITHM:
        raise ValueError(f"invalid checksum algorithm {algorithm}")

    return algorithm, checksum


HEXDIGEST_RE = re.compile(r"^[0-9a-fA-F]+$")


def is_hexdigest(algorithm: str, value: str) -> bool:
    return bool(
        len(value) == HASH_ALGORITHM_LENGTH[algorithm] and HEXDIGEST_RE.search(value)
    )


def find_checksum_in_file(content: str, filename: str) -> str | None:
    lines = content.splitlines()
    for line in lines:
        match = re.search(r"^([0-9a-fA-F]+)\s+" + re.escape(filename) + r"$", line)
        if match is not None:
            return match.group(1)

    return None


def compute_file_checksum(algorithm: str, filepath: Path) -> str:
    mixer = hashlib.new(algorithm, usedforsecurity=False)

    with filepath.open("rb") as file:
        while True:
            blob = file.read(8192)
            if not blob:
                break
            mixer.update(blob)

    digest = mixer.hexdigest()
    logger.debug("Computed %s digest '%s'", algorithm, digest)

    return digest



from __future__ import annotations

import logging
import sys
from os import environ
from pathlib import Path
from shutil import move, unpack_archive
from tempfile import TemporaryDirectory

from requests import Session

from .checksum import (
    compute_file_checksum,
    find_checksum_in_file,
    is_hexdigest,
    parse_checksum_option,
)
from .unpack import register_unpack_formats

__all__ = ["GhReleaseInstall"]

LATEST = "latest"

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# pylint: disable=too-few-public-methods
class Release:
    def __init__(self, tag: str) -> None:
        self.tag = tag

    @property
    def version(self) -> str:
        return self.tag.strip("v")


def get_latest_tag(session: Session, repository: str) -> str:
    url = f"https://api.github.com/repos/{repository}/releases/latest"
    with session.get(url) as res:
        res.raise_for_status()
        body = res.json()

    return body["tag_name"]


# pylint: disable=too-many-instance-attributes
class GhReleaseInstall:
    _target: Release | None = None
    _local: Release | None = None
    _session: Session

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        repository: str,
        asset: str,
        destination: str | Path,
        extract: str | None = None,
        version: str = LATEST,
        version_file: str | None = None,
        checksum: str | None = None,
    ):
        self._repository = repository
        self._asset = asset
        self._destination = str(destination)
        self._extract = extract
        self._version = version
        self._version_file = version_file

        self.checksum_algorithm, self.checksum = None, None
        if checksum is not None:
            self.checksum_algorithm, self.checksum = parse_checksum_option(checksum)

        self._session = Session()
        if "GITHUB_TOKEN" in environ:
            logger.debug("Loading GITHUB_TOKEN from env")
            github_token = environ.get("GITHUB_TOKEN")
            self._session.headers.update({"Authorization": f"token {github_token}"})

        register_unpack_formats()

    def _resolve_path(self, path: str, **variables: str) -> str:
        if self._target is not None:
            variables["tag"] = self._target.tag
            variables["version"] = self._target.version

        return path.format(**variables)

    @property
    def asset(self) -> str:
        return self._resolve_path(self._asset)

    @property
    def destination(self) -> Path:
        destination = Path(self._resolve_path(self._destination))

        if destination.is_dir():
            return destination / self.asset

        return destination

    @property
    def extract(self) -> str | None:
        if self._extract is None:
            return None
        return self._resolve_path(self._extract)

    @property
    def version_file(self) -> Path | None:
        if self._version_file is None:
            return None

        return Path(
            self._resolve_path(
                self._version_file,
                destination=str(self.destination),
            )
        )

    def _github_asset_url(self, asset: str) -> str:
        assert self._target is not None
        return f"https://github.com/{self._repository}/releases/download/{self._target.tag}/{asset}"

    def _get_target_version(self):
        """
        If not provided, get latest tag/version from the Github repository.
        """
        if self._version == LATEST:
            self._target = Release(get_latest_tag(self._session, self._repository))
        else:
            self._target = Release(self._version)

        logger.debug("Target version is '%s'", self._target.version)

    def _get_local_version(self):
        """
        Get local tag / version from possible version file.
        """
        if self.version_file is not None and self.version_file.is_file():
            self._local = Release(self.version_file.read_text(encoding="utf-8"))
            logger.debug("Local version is '%s'", self._local.version)

    def _get_checksum_from_url(self, url: str) -> str | None:
        """
        Download checksum file from the provided url and extract the checksum.
        """
        with self._session.get(url) as res:
            if res.status_code == 404:
                return None
            res.raise_for_status()

            return find_checksum_in_file(res.text, self.asset)

    def _verify_checksum(self, asset_file: Path) -> bool:
        """
        Verify asset checksum, first check against a possible hand written digest,
        then check against a digest from a asset checksum file.
        """
        assert self.checksum is not None
        assert self.checksum_algorithm is not None

        local_checksum = compute_file_checksum(self.checksum_algorithm, asset_file)

        # We hope nobody will ever pass a asset filename that matches this check
        if is_hexdigest(self.checksum_algorithm, self.checksum):
            return local_checksum == self.checksum

        target_checksum_url = self._github_asset_url(self.checksum)
        target_checksum = self._get_checksum_from_url(target_checksum_url)
        return local_checksum == target_checksum

    def _download_release_asset(self, tmp_dir: Path):
        """
        Download target version release file in a temporary file.
        """
        url = self._github_asset_url(self.asset)
        with self._session.get(url, stream=True) as res:
            res.raise_for_status()
            tmp_file = tmp_dir / self.asset

            logger.debug("Saving asset to '%s'", tmp_file)
            with tmp_file.open("wb") as tmp_fd:
                for chunk in res.iter_content(chunk_size=2048):
                    tmp_fd.write(chunk)

        return tmp_file

    def _extract_release_asset(self, tmp_dir: Path, asset_file: Path) -> Path:
        """
        Extract downloaded release archive.
        """
        unpack_archive(asset_file, tmp_dir)
        assert self.extract is not None
        return tmp_dir / self.extract

    def run(self):
        self._get_target_version()
        self._get_local_version()

        if self._local is not None:
            if not self.destination.is_file():
                logger.warning(
                    "Local version is referring to an inexistent asset '%s'",
                    self.destination,
                )
            elif self._target.version == self._local.version:
                logger.info("Target version is already installed")
                sys.exit(0)

        with TemporaryDirectory(prefix="gh-release-installer") as tmp_dir:
            tmp_dir = Path(tmp_dir)
            asset_file = self._download_release_asset(tmp_dir)

            if self.checksum is not None:
                if not self._verify_checksum(asset_file):
                    logger.error("Checksum verification failed")
                    sys.exit(1)
                logger.info("Checksum verification succeeded")

            if self.extract is not None:
                asset_file = self._extract_release_asset(tmp_dir, asset_file)
                logger.info("Extracted archive to '%s'", asset_file)

            move(asset_file, self.destination)
            self.destination.chmod(0o755)
            logger.info("Installed file to '%s'", self.destination)

        # Save to local tag/version file
        if self.version_file is not None:
            self.version_file.write_text(self._target.tag, encoding="utf-8")
            logger.info("Saved version file to '%s'", self.version_file)



from __future__ import annotations

import logging
import sys
from argparse import (
    ArgumentDefaultsHelpFormatter,
    ArgumentParser,
    RawDescriptionHelpFormatter,
)

from gh_release_install import GhReleaseInstall
from gh_release_install.checksum import HASH_ALGORITHM

logger = logging.getLogger(__name__)


class ArgumentParserFormatter(
    RawDescriptionHelpFormatter,
    ArgumentDefaultsHelpFormatter,
):
    pass


parser = ArgumentParser(
    description="Install GitHub release file on your system.",
    formatter_class=lambda prog: ArgumentParserFormatter(prog, width=80),
)
parser.add_argument(
    "repository",
    metavar="REPOSITORY",
    help="Github REPOSITORY org/repo to get the release from.",
)
parser.add_argument(
    "asset",
    metavar="ASSET",
    help="Release ASSET filename. May contain variables such as '{version}' or '{tag}'.",
)
parser.add_argument(
    "--extract",
    metavar="<filename>",
    help="""Extract the <filename> from the release asset archive and install the
            extracted file instead. May contain variables such as '{version}' or
            '{tag}'.""",
)
parser.add_argument(
    "destination",
    metavar="DESTINATION",
    help="""Path to save the downloaded file. If DESTINATION is a directory, the asset
            name will be used as filename in that directory. May contain variables such
            as '{version}' or '{tag}'.""",
)
parser.add_argument(
    "--version",
    default="latest",
    metavar="<version>",
    help="""Desired release version to install. When using 'latest' the installer will
            guess the latest version from the Github API.""",
)
parser.add_argument(
    "--version-file",
    metavar="<filename>",
    help="""Track the version installed on the system using a file. May contain
            variables such as '{destination}'.""",
)
parser.add_argument(
    "--checksum",
    metavar="<hash>:<digest|asset>",
    help=f"""Asset checksum used to verify the downloaded ASSET. <hash> can be one of
             {', '.join(HASH_ALGORITHM)}. <digest|asset> can either be the expected
             checksum, or the filename of an checksum file in the release assets.""",
)
parser.add_argument(
    "-v",
    "--verbose",
    dest="verbosity",
    action="count",
    default=0,
    help="Increase the verbosity.",
)
parser.add_argument(
    "-q",
    "--quiet",
    dest="verbosity",
    action="store_const",
    const=-1,
    help="Disable logging.",
)
parser.epilog = """
template variables:
    {tag}               Release tag name.
    {version}           Release tag name without leading 'v'.
    {destination}       DESTINATION path, including the asset filename if path
                        is a directory.

examples:
    gh-release-install 'mvdan/sh' \\
        'shfmt_{tag}_linux_amd64' \\
        '/usr/local/bin/shfmt' \\
        --version 'v3.3.1'

    gh-release-install 'prometheus/prometheus' \\
        'prometheus-{version}.linux-amd64.tar.gz' \\
        --extract 'prometheus-{version}.linux-amd64/prometheus' \\
        '/usr/local/bin/prometheus' \\
        --version-file '{destination}.version' \\
        --checksum 'sha256:sha256sums.txt'
"""


def run():
    args = parser.parse_args()

    if args.verbosity is not None and args.verbosity >= 0:
        levels = [logging.ERROR, logging.INFO, logging.DEBUG]
        logging.basicConfig(
            level=levels[min(args.verbosity, 2)],
            format="%(levelname)s:\t%(message)s",
        )

    installer = GhReleaseInstall(
        repository=args.repository,
        asset=args.asset,
        destination=args.destination,
        extract=args.extract,
        version=args.version,
        version_file=args.version_file,
        checksum=args.checksum,
    )

    try:
        installer.run()
    # pylint: disable=broad-except
    except Exception as exception:
        logger.exception(exception)
        sys.exit(1)
